#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

// #define DEBUG_LOG

#define HI(n) ((uint64_t)(n)>>32)
#define LO(n) ((uint64_t)(n)&0xffffffff)
#define U128(high,low) ((__uint128_t)(uint64_t)high<<64 | (uint64_t)low)

//---------------------------------------------------------------------------
// 0.45.0 - known values
//---------------------------------------------------------------------------

uint64_t magic_table[16] = {
  0x2dd7caaefcf073eb, 0xa9209937349cfe9c,
  0xb84bfc934b0e60ef, 0xff709c157b26e477,
  0x3936fd8735455112, 0xca141bf22338d331,
  0xdd40e749cb64fd02, 0x5e268f564b0deb26,
  0x658239596bdea9ec, 0x31cedf33ac38c624,
  0x12f56816481b0cfd, 0x94e9de155f40f095,
  0x5089c907844c6325, 0xdf887e97d73c50e3,
  0xae8870787ce3c11d, 0xa6767d18c58d2117,
};
#define ROUND_MAGIC_HI 0xe3f0d44988bcdfab
#define ROUND_MAGIC_LO 0x081570afdd535ec3
#define FINAL_MAGIC0 0xce7c4801d683e824
#define FINAL_MAGIC1 0x6823775b1daad522
#define ROUND_MAGIC U128(ROUND_MAGIC_HI,ROUND_MAGIC_LO)
// -- end

//---------------------------------------------------------------------------

// PHASE 1: 
// - find the 8 combinations for ROUND_MAGIC, FINAL_MAGIC0 and FINAL_MAGIC1
uint32_t *magic_candidates;
int       magic_candidate_count = 0;

// global iterators
int g_i0, g_i1, g_i2, g_i3, g_i4, g_i5, g_i6, g_i7;

// functions for brute-force
uint64_t get_ROUND_MAGIC_HI()
{
//return ROUND_MAGIC_HI;
  return ((uint64_t)magic_candidates[g_i1] << 32) | magic_candidates[g_i0];
}

uint64_t get_ROUND_MAGIC_LO()
{
//return ROUND_MAGIC_LO;
  return ((uint64_t)magic_candidates[g_i3] << 32) | magic_candidates[g_i2];
}

uint64_t get_FINAL_MAGIC0()
{
//return FINAL_MAGIC0;
  return ((uint64_t)magic_candidates[g_i5] << 32) | magic_candidates[g_i4];
}

uint64_t get_FINAL_MAGIC1()
{
//return FINAL_MAGIC1;
  return ((uint64_t)magic_candidates[g_i7] << 32) | magic_candidates[g_i6];
}

__uint128_t get_ROUND_MAGIC()
{
  return U128(get_ROUND_MAGIC_HI(),get_ROUND_MAGIC_LO());
}

// END
//---------------------------------------------------------------------------

static __uint128_t hash_muladd(__uint128_t, __uint128_t, __uint128_t);
static __uint128_t hash_chunk(const uint8_t *, int64_t);
static uint64_t read_int64(const uint8_t *);

uint64_t read_int64(const uint8_t *p)
{
  int i;
  // endian-safe read 64-bit integer
  uint64_t n = 0;
  for (i = 7; i >= 0; i--) {
    n = (n << 8) | p[i];
  }
  return n;
}

uint64_t Hash(uint8_t *in, uint32_t len)
{
  uint32_t num_chunks = len / 128;

  // copy tail, pad with zeroes
  uint8_t tail[128] = {0};
  int tail_size = len % 128;
  memcpy(tail, in + len - tail_size, tail_size);

  __uint128_t hash;
  if (num_chunks) {
    // Hash the first 128 bytes
    hash = hash_chunk(in, 128);

  } else {
    // Hash the tail
    hash = hash_chunk(tail, tail_size);
  }

  hash += get_ROUND_MAGIC();

  if (num_chunks) {
    while (--num_chunks) {
      in += 128;
      hash = hash_muladd(hash, get_ROUND_MAGIC(), hash_chunk(in, 128));
    }

    if (tail_size) {
      hash = hash_muladd(hash, get_ROUND_MAGIC(), hash_chunk(tail, tail_size));
    }
  }

  // Finalize the hash
  hash += U128(tail_size * 8, 0);
  if (hash >= U128(0x7fffffffffffffff,0xffffffffffffffff)) {
    hash++;
  }
  hash = hash << 1 >> 1;

  uint64_t hash_high = hash >> 64;
  uint64_t hash_low = hash;
  uint64_t X = hash_high + HI(hash_low);
  X = HI(X + HI(X) + 1) + hash_high;
  uint64_t Y = (X << 32) + hash_low;

  uint64_t A = X + get_FINAL_MAGIC0();
  if (A < X) {
    A += 0x101;
  }

  uint64_t B = Y + get_FINAL_MAGIC1();
  if (B < Y) {
    B += 0x101;
  }

  __uint128_t H = (__uint128_t) A * B;
  H = 0x101 * (H >> 64) + (uint64_t) H;
  H = 0x101 * (H >> 64) + (uint64_t) H;
  if (H >> 64) {
    H += 0x101;
  }
  if ((uint64_t) H > 0xFFFFFFFFFFFFFEFE) {
    H += 0x101;
  }
  return (uint64_t) H;
}

__uint128_t hash_chunk(const uint8_t *chunk, int64_t size)
{
  int i;
  __uint128_t hash = 0;
  for (i = 0; i < 8; i++) {
    int offset = i * 16;
    if (offset >= size) {
      break;
    }
    uint64_t a = read_int64(chunk + offset);
    uint64_t b = read_int64(chunk + offset + 8);
    hash += (__uint128_t) (a + magic_table[i * 2]) *
      (__uint128_t) (b + magic_table[i * 2 + 1]);
  }
  return hash << 2 >> 2;
}

__uint128_t hash_muladd(__uint128_t hash, __uint128_t mul, __uint128_t add)
{
  uint64_t a0 = LO(add), a1 = HI(add), a23 = add >> 64;
  uint64_t m0 = LO(mul),        m1 = HI(mul);
  uint64_t m2 = LO(mul >> 64),  m3 = HI(mul >> 64);
  uint64_t h0 = LO(hash),       h1 = HI(hash);
  uint64_t h2 = LO(hash >> 64), h3 = HI(hash >> 64);

  /* Column sums, before carry */
  uint64_t c0 = (h0 * m0);
  uint64_t c1 = (h0 * m1) + (h1 * m0);
  uint64_t c2 = (h0 * m2) + (h1 * m1) + (h2 * m0);
  uint64_t c3 = (h0 * m3) + (h1 * m2) + (h2 * m1) + (h3 * m0);
  uint64_t c4 = (h1 * m3) + (h2 * m2) + (h3 * m1);
  uint64_t c5 = (h2 * m3) + (h3 * m2);
  uint64_t c6 = (h3 * m3);

  /* Combine, add, and carry (bugs included) */
  uint64_t r2 = c2 + (c6 << 1) + a23;
  uint64_t r3 = c3                   + HI(r2);
  uint64_t r0 = c0 + (c4 << 1) + a0  + (r3 >> 31);
  uint64_t r1 = c1 + (c5 << 1) + a1  + HI(r0);

  /* Return as uint128_t */
  __uint128_t result = ((r3 << 33 >> 1) | LO(r2)) + HI(r1);
  return (result << 64) | (r1 << 32) | LO(r0);
}

int main(int argc, char **argv)
{
  FILE *f;
  char  line[64], *p, *_p;
  int   i, min, max;

  fprintf(stdout, "PogoHash-brute #1\n");
  fprintf(stdout, "-----------------\n");
  fprintf(stdout, "\n");

  // load our candidates
  f = fopen("x.candidates", "r");
  i = 0;
  while (fgets(line, 64, f) != NULL) 
  {
    p = line;
    // get the number of candidates
    if (magic_candidate_count == 0)
    {
      magic_candidate_count = strtol(p, &_p, 10);
      magic_candidates = malloc(magic_candidate_count * sizeof(uint32_t));
    }
    else
      magic_candidates[i++] = strtol(&p[2], &_p, 16);
  }
  fclose(f);

  // lets find 0xfffffefe : our magic number
  min = 0;
  max = magic_candidate_count-1;
  while ((max > 8) && (magic_candidates[max] != 0xfffffefe)) max--;

  // we know that the value isn't a magic
  max--;
 
  // diff between x.len_0.load and x.len_1.load happens at index 57
  min = 57;

//max = 106, min = 57 :: real	0m2.134s
//max = 107, min = 57 :: real	1m12.668s
//max = 108, min = 57 :: real	2m42.808s
//max = 109, min = 57 :: real	4m34.226s
//max = 110, min = 57 :: real	6m50.485s
//max = 119, min = 57 :: real	57m0.064s <<-- this is based on logic

  fprintf(stdout, "x.candidates\n");
  fprintf(stdout, "------------\n");
  fprintf(stdout, "  total: %d\n", magic_candidate_count);
  fprintf(stdout, "  min:   %d\n", min);
  fprintf(stdout, "  max:   %d\n", max);
  fprintf(stdout, "\n");
 
  fprintf(stdout, ":: Hash(buffer, sizeof(buffer) [iOS code]\n");
  {
    unsigned long result;
    uint8_t buffer[256] = {0};
    int j;

    // this is our test vector
    j = 0;

    //
    // BRUTE FORCE
    //

    fprintf(stdout, "scanning:\n");

    // FINAL_MAGIC1
    g_i7 = max;
loop_i7:
    {
      fprintf(stdout, "g_i7=%d\n", g_i7);

      g_i6 = max;
loop_i6:
      {
        // special case: cannot happen
        if (g_i6 == g_i7) goto loop_i6_done; 

        // we must find the lowest of the two indexes
        int t1 = (g_i6 > g_i7) ? g_i7 : g_i6;

        // FINAL_MAGIC0
        g_i5=t1-1;
loop_i5:
        {
          g_i4=t1-1;
loop_i4:
          {
            // special case: cannot happen
            if (g_i4 == g_i5) goto loop_i4_done;

            // we must find the lowest of the two indexes
            int t2 = (g_i4 > g_i5) ? g_i5 : g_i4;

            // ROUND_MAGIC
            g_i3=t2-1; 
loop_i3: 
            {
              g_i2=t2-1;
loop_i2:
              {
                // special case: cannot happen
                if (g_i2 == g_i3) goto loop_i2_done;

                g_i1=t2-1;
loop_i1:
                {
                  // special case: cannot happen
                  if ((g_i1 == g_i2) || (g_i1 == g_i3)) goto loop_i1_done;

                  g_i0=t2-1;
loop_i0:
                  {
                    // special case: cannot happen
                    if ((g_i0 == g_i1) || (g_i0 == g_i2) || (g_i0 == g_i3))
                      goto loop_i0_done;

                    result = Hash(buffer, j);
#ifdef DEBUG_LOG
                    fprintf(stdout, ":: %2d %2d %2d %2d %2d %2d %2d %2d\n", 
                            g_i0, g_i1, g_i2, g_i3, g_i4, g_i5, g_i6, g_i7);
#endif

                    // have we found them?
                    if (result == 0xeebb9c527a4d5f53) 
                    {
                      fprintf(stdout, ":: %2d %2d %2d %2d %2d %2d %2d %2d\n", 
                              g_i0, g_i1, g_i2, g_i3, g_i4, g_i5, g_i6, g_i7);

                      for (i=0; i<j; i++)
                        fprintf(stdout, "%02x ", (buffer[i] & 0xff));
                      fprintf(stdout, "len=%2d\n", j);
                      printf("0x%08lx\n", result);

                      goto MAGICS_FOUND;
                    }

loop_i0_done:
                    g_i0--;
                  }
                  if (g_i0>=min) goto loop_i0;

loop_i1_done:
                  g_i1--;
                }
                if (g_i1>=min) goto loop_i1;

loop_i2_done:
                g_i2--;
              }
              if (g_i2>=min) goto loop_i2;

              g_i3--;
            } 
            if (g_i3>=min) goto loop_i3;
          
loop_i4_done: 
            g_i4--;
          }
          if (g_i4 >= (min+2)) goto loop_i4;

          g_i5--;
        }
        if (g_i5 >= (min+2)) goto loop_i5;

loop_i6_done:
        g_i6--;
      } 
      if (g_i6 >= (min+6)) goto loop_i6;

      g_i7--;
    } 
    if (g_i7 >= (min+6)) goto loop_i7;
  }

  fprintf(stdout, "NOT FOUND\n");
  goto MAGICS_NOT_FOUND;

MAGICS_FOUND:;

  fprintf(stdout, "ROUND_MAGIC_HI = 0x%08x%08x\n", 
                  magic_candidates[g_i1], magic_candidates[g_i0]);
  fprintf(stdout, "ROUND_MAGIC_LO = 0x%08x%08x\n", 
                  magic_candidates[g_i3], magic_candidates[g_i2]);
  fprintf(stdout, "FINAL_MAGIC0   = 0x%08x%08x\n", 
                  magic_candidates[g_i5], magic_candidates[g_i4]);
  fprintf(stdout, "FINAL_MAGIC1   = 0x%08x%08x\n", 
                  magic_candidates[g_i7], magic_candidates[g_i6]);

MAGICS_NOT_FOUND:

  // clean up
  if (magic_candidates != NULL) free(magic_candidates);

  fprintf(stdout, "done\n");
  return 0;
}
